import socket
import multiprocessing
import logging
from typing import Tuple, Dict
import signal
import sys

from bank_protocol.command_handler import CommandHandler
from core import Request, Response
from core.exceptions import BankNodeError


class BankWorker(multiprocessing.Process):
    def __init__(self, client_socket: socket.socket, client_address: Tuple, config: Dict):
        super().__init__()

        self.logger = logging.getLogger(__name__)

        self.client_socket = client_socket
        self.client_socket.settimeout(config["client_idle_timeout"])
        self.client_address = client_address

        self.command_handler = CommandHandler(config)
        self.config = config

    def __setup_signals(self):
        self.logger.debug("Setting up exit signal hooks for worker")

        # Handle C standard signals
        signal.signal(signal.SIGTERM, self.gracefully_exit_worker)
        signal.signal(signal.SIGINT, self.gracefully_exit_worker)

        # Handle windows related signals
        if sys.platform == "win32":
            signal.signal(signal.CTRL_C_EVENT, self.gracefully_exit_worker)
            signal.signal(signal.CTRL_BREAK_EVENT, self.gracefully_exit_worker)
            signal.signal(signal.SIGBREAK, self.gracefully_exit_worker)

    def run(self):
        self.__setup_signals()
        with self.client_socket:
            while True:
                try:
                    raw_request = self.client_socket.recv(1024).decode("utf-8")

                    if not raw_request:
                        self.logger.debug("%s disconnected", self.client_address[0])
                        break

                    request = Request(raw_request)
                    self.logger.debug("Received request from %s - %s", self.client_address[0], request)

                    response: Response = self.command_handler.execute(request) + "\n\r"

                    self.client_socket.sendall(response.encode("utf-8"))

                except socket.timeout:
                    self.logger.debug("Client was idle for too long. Ending connection")
                    response = "ER Idle too long\n\r"
                    self.client_socket.sendall(response.encode("utf-8"))
                    self.client_socket.shutdown(socket.SHUT_RDWR)
                    self.client_socket.close()
                    break
                except BankNodeError as e:
                    response = "ER " + e.message + "\n\r"
                    self.client_socket.sendall(response.encode("utf-8"))
                except socket.error as e:
                    response = "ER Internal server error\n\r"
                    self.logger.error(e)
                    break

        self.logger.debug("Closing process for %s", self.client_address[0])

    def gracefully_exit_worker(self, signum, _):
        """Log the signal caught and exit with status 0"""

        signal_name = signal.Signals(signum).name
        self.logger.warning("Worker caught %s - cleaning up", signal_name)
        self.cleanup()
        sys.exit(0)

    def cleanup(self):
        self.logger.info("Closing connection with %s", self.client_address[0])
        if self.client_socket:
            self.client_socket.close()


__all__ = ["BankWorker"]
